#
# Copyright (c) 2010 Nokia Corporation and/or its subsidiary(-ies).
# All rights reserved.
# This component and the accompanying materials are made available
# under the terms of "Eclipse Public License v1.0"
# which accompanies this distribution, and is available
# at the URL "http://www.eclipse.org/legal/epl-v10.html".
#
# Initial Contributors:
# Nokia Corporation - initial contribution.
#
# Contributors:
#
# Description:
# Read generic link info XML generated by SbsLinkInfoReader, create BuildData.
#

''' Read generic link info XML generated by SbsLinkInfoReader, create BuildData. '''

import xml.sax
import os
import sys

from Blocks.Packaging.BuildData import PlainBuildData, BdFile
from Blocks.Packaging.Logging import Logging

class LinkInfo(object):
    '''
    Created from LinkInfo XML output. Used to add dependencies to a buildData
    instance.

    Usage::
        lf = LinkInfo(xml)
        lf.addDependencies(myBuildData)

    Doctest::
        >>> import StringIO
        >>> xml = StringIO.StringIO("""<linkdata epocroot="/data/dragonfly/workspaces/1733/">
        ...     <inf path="/data/dragonfly/workspaces/1733/src/common/generic/graphics/egl/group/bld.inf" ddir="foo/bar">
        ...       <target path="epoc32/release/armv5/udeb/libegl.dll" component="EGL Implementation" platform="armv5" dso="">
        ...         <libraries>
        ...           <lib>epoc32/release/armv5/lib/dfpaeabi.dso</lib>
        ...           <lib>epoc32/release/armv5/lib/dfprvct2_2.dso</lib>
        ...           <lib>epoc32/release/armv5/lib/drtrvct2_2.dso</lib>
        ...         </libraries>
        ...       </target>
        ...     </inf>
        ... </linkdata>""")
        >>> lf = LinkInfo(xml)
        >>> data = PlainBuildData()
        >>> data.addTargetFiles(['epoc32/release/armv5/udeb/libegl.dll'])
        >>> data = lf.addDependencies(data)
        >>> assert len(data.getDependencies()[0].getOwnerDependencies()) == 3
        >>> assert len(data.getDependencies()[0].getSourceDependencies()) == 3
        >>> assert data.dependencyData["DotDeeDependencyProcessor"] == {"ddirs": ["foo/bar"]}
    '''

    def __init__(self, linkInfoXml):
        '''
        @param linkInfoXml: Link info XML
        @type linkInfoXml: Filename or file-like object
        '''
        self.reader = LinkInfoXmlReader()
        self.dependencies = self.reader.getBuildData(linkInfoXml)

    def getNonstandardInterfaces(self):
        '''
        Get nonstandard interfaces extracted from link info

        @return: Mapping of binary to correct api file (.dso)
        @rtype: Dictionary(String)
        '''
        apiMapping = {}
        for bd in self.dependencies.itervalues():
            apiMapping.update(bd.apiMapping)
        return apiMapping

    def addDependencies(self, buildData, bldInfPath=None, newTarget=None):
        '''
        Add dependencies to buildData. If bldInfPath is not given, the entire
        LinkInfo is searched for matching files and newTarget is ignored.

        Also adds dependencyData["DotDeeDependencyProcessor"]["ddirs"]

        @param buildData: A BuildData object to add dependencies to
        @type buildData: BuildData
        @param bldInfPath: Optional absolute path to bldInf if it is known. Speeds up the operation.
        @type bldInfPath: String
        @param newTarget: What to do with new targets from the link info, one of "add", "log" or anything else to ignore
        @type newTarget: String or None
        '''
        if bldInfPath:
            bldInfPath = os.path.normpath(bldInfPath)
            if bldInfPath in self.dependencies:
                return self._addDeps(buildData, bldInfPath, newTarget)
        else:
            for bldInfPath in self.dependencies:
                self._addDeps(buildData, bldInfPath, newTarget=None)

    def _addDeps(self, buildData, bldInfPath, newTarget=None):
        '''
        Add dependencies to buildData.

        @param buildData: A BuildData object to add dependencies to
        @type buildData: BuildData
        @param bldInfPath: absolute path to bldInf, speeds up the operation
        @type bldInfPath: String
        @param newTarget: What to do with new targets from the link info, one of "add", "log" or or anything else to ignore
        @type newTarget: String or None
        '''
        for bdFile in self.dependencies[bldInfPath].getDependencies():
            if bdFile.getPath() in buildData.getTargetFiles():
                # no dependency data above, only paths - OK to overwrite
                buildData.addDeliverable(bdFile)
            elif newTarget == "add":
                buildData.addDeliverable(bdFile)
            elif newTarget == "log":
                Logging.getLogger("pfw.datasources.linkinfo").warning(
                    "Link data from %s target %s does not exist in %s", bldInfPath, bdFile.getPath(), buildData.getComponentName())

            # Don't overwrite existing dependencyData, add any missing bits
            for d in self.dependencies[bldInfPath].dependencyData:
                if not buildData.dependencyData.get(d):
                    buildData.dependencyData[d] = self.dependencies[bldInfPath].dependencyData[d]
                else:
                    for k, v in self.dependencies[bldInfPath].dependencyData[d].items():
                        if not buildData.dependencyData.get(d).get(k):
                            buildData.dependencyData[d][k] = v

class LinkInfoXmlReader(xml.sax.ContentHandler):
    '''
    Process linkdata:
        - inf -> BuildData
        - target path -> BdFile
        - libraries -> indirectDependencies

    Sample linkdata::
        <linkdata epocroot="/data/dragonfly/workspaces/1733/">
            <inf path="/data/dragonfly/workspaces/1733/src/common/generic/graphics/egl/group/bld.inf" ddir="foo/bar">
              <target path="epoc32/release/armv5/udeb/libegl.dll" component="EGL Implementation" platform="armv5" dso="">
                <libraries>
                  <lib>epoc32/release/armv5/lib/euser.dso</lib>
                  <lib>epoc32/release/armv5/lib/estlib.dso</lib>
                  <lib>epoc32/release/armv5/lib/ws32.dso</lib>
                  <lib>epoc32/release/armv5/lib/cone.dso</lib>
                  <lib>epoc32/release/armv5/lib/bitgdi.dso</lib>
                  <lib>epoc32/release/armv5/lib/fbscli.dso</lib>
                  <lib>epoc32/release/armv5/lib/hal.dso</lib>
                  <lib>epoc32/release/armv5/lib/gdi.dso</lib>
                  <lib>epoc32/release/armv5/lib/palette.dso</lib>
                  <lib>epoc32/release/armv5/lib/drtaeabi.dso</lib>
                  <lib>epoc32/release/armv5/lib/dfpaeabi.dso</lib>
                  <lib>epoc32/release/armv5/lib/dfprvct2_2.dso</lib>
                  <lib>epoc32/release/armv5/lib/drtrvct2_2.dso</lib>
                </libraries>
              </target>
              <target path="epoc32/release/armv5/urel/libegl.dll" component="EGL Implementation" platform="armv5" dso="">
                <libraries>
                  <lib>epoc32/release/armv5/lib/euser.dso</lib>
                  <lib>epoc32/release/armv5/lib/estlib.dso</lib>
                  <lib>epoc32/release/armv5/lib/ws32.dso</lib>
                  <lib>epoc32/release/armv5/lib/cone.dso</lib>
                  <lib>epoc32/release/armv5/lib/bitgdi.dso</lib>
                  <lib>epoc32/release/armv5/lib/fbscli.dso</lib>
                  <lib>epoc32/release/armv5/lib/hal.dso</lib>
                  <lib>epoc32/release/armv5/lib/gdi.dso</lib>
                  <lib>epoc32/release/armv5/lib/palette.dso</lib>
                  <lib>epoc32/release/armv5/lib/drtaeabi.dso</lib>
                  <lib>epoc32/release/armv5/lib/dfpaeabi.dso</lib>
                  <lib>epoc32/release/armv5/lib/dfprvct2_2.dso</lib>
                  <lib>epoc32/release/armv5/lib/drtrvct2_2.dso</lib>
                </libraries>
              </target>
            </inf>
        </linkdata>
    '''
    def __init__(self):
        xml.sax.ContentHandler.__init__(self)
        self.logger = Logging.getLogger("pfw.datasources.linkinforeader")

    @staticmethod
    def getBuildData(source):
        linkReader = LinkInfoXmlReader()
        p = xml.sax.make_parser()
        p.setContentHandler(linkReader)
        if hasattr(source, "read"):
            file = source
        else:
            file = open(source)
        while True:
            data = file.read(1000000)
            if not data:
                break
            p.feed(data)
        p.close()
        return linkReader.infAndBuildData

    def characters(self, data):
        self.cdata.append(data)

    def startDocument(self):
        self.infAndBuildData = {}
        self.epocroot = None
        # currently handled item
        self.cdString = ""
        self.buildData = None
        self.inf = None
        self.bdFile = None

    def startElement(self, tag, attributes):
        self.cdata = []
        self.cdString = ""
        if tag == "linkdata":
            self.epocroot = attributes.get("epocroot")
        elif tag == "inf":
            self.inf = os.path.normpath(attributes.get("path"))
            self.buildData = PlainBuildData()
            self.buildData.setTargetRoot(self.epocroot)
            self.buildData.dependencyData["DotDeeDependencyProcessor"] = {"ddirs": [attributes.get("ddir")]}
        elif tag == "target":
            path = attributes.get("path")
            self.bdFile = BdFile(path)
            self.bdFile.variantPlatform = attributes.get("platform")
            dso = attributes.get("dso")
            if dso:
                self.buildData.addNonstandardInterfaces({path: dso})
            if attributes.get("component"):
                # Not used for component naming but store it anyway.
                if self.buildData.getComponentName():
                    if self.buildData.getComponentName() != attributes.get("component"):
                        # Multiple components from same INF possible?
                        # Hmm ... got to watch for this if we ever use component names from this output
                        self.logger.info("Targets from same INF %s have different component names: %s vs. %s",
                            self.inf, self.buildData.getComponentName(), attributes.get("component"))
                else:
                    self.buildData.setComponentName(attributes.get("component"))

    def endElement(self, tag):
        self.cdString = "".join(self.cdata).strip()
        if tag == "inf":
            if self.inf not in self.infAndBuildData:
                self.infAndBuildData[self.inf] = []
            self.infAndBuildData[self.inf] = self.buildData
            self.buildData = None
        elif tag == "target":
            self.buildData.addDeliverable(self.bdFile)
            self.bdFile = None
        elif tag == "lib":
            # dependency on epoc32/release/armv5/lib/drtrvct2_2.dso
            # -> source of bdFile depends on owner of drtrvct2_2.dso
            self.bdFile.addSourceDependency(self.cdString)
            # -> owner of bdFile depends on epoc32/release/armv5/urel|udeb/drtrvct2_2.dll
            dep = os.path.join("epoc32/release/",
                    self.bdFile.getVariantPlatform(),
                    self.bdFile.getVariantType(),
                    os.path.basename(self.cdString).split(".dso")[0] + ".dll")
            self.bdFile.addOwnerDependency(dep)

if __name__ == "__main__":
    import doctest
    doctest.testmod(sys.modules[__name__])