#
# Copyright (c) 2010 Nokia Corporation and/or its subsidiary(-ies).
# All rights reserved.
# This component and the accompanying materials are made available
# under the terms of "Eclipse Public License v1.0"
# which accompanies this distribution, and is available
# at the URL "http://www.eclipse.org/legal/epl-v10.html".
#
# Initial Contributors:
# Nokia Corporation - initial contribution.
#
# Contributors:
#
# Description:
# Read build log output using two alternative methods
#

'''
Read build log output using two alternative methods::
    - WhatLogReader reads current SBS (2.4.2) output
    - GenericLogReader reads a more generic format converted from SBS output
    using XSL.

The collected information is::
    - INF file path
    - Target files generated
    - Target file types (resource|build|bitmap|member)

    A member is a file stored in an archive. The parent archive can be obtained
    via WhatLogReader.getFullInfo()

Examples are included at the end in the command line tool.
'''

import os
import sys
import xml.sax

from Blocks.Packaging.BuildData import PlainBuildData
from Blocks.Packaging.Logging import Logging

class WhatLogReader(xml.sax.ContentHandler):
    '''
    Read output from makefiles created using::

        sbs -c <foo>.whatlog

    The build log (namespace http://symbian.com/xml/build/log) with interesting
    parts italicized::

        <buildinfo>
            <whatlog bldinf='I{/path/to/src/path/to/group/bld.inf}' mmp='' config='foo.whatlog'>
                <export destination='I{/path/to/epoc32/foo/bar}' source='/path/to/src/path/to/baz'/>
            </whatlog>

            <whatlog bldinf='I{/path/to/src/path/to/group/bld.inf}' mmp='' config='foo.whatlog'>
                <archive zipfile='/path/to/src/path/to/foo.zip'>
                    <member>I{/path/to/epoc32/foo/bar}</member>
                </archive>
            </whatlog>

            <whatlog bldinf='I{/path/to/src/path/to/group/bld.inf}' mmp='' config='foo.whatlog'>
                <resource>I{foo/bar}</resource>
            </whatlog>

            <whatlog bldinf='I{/path/to/src/path/to/group/bld.inf}' mmp='' config='foo.whatlog'>
                <build>I{foo/bar}</build>
            </whatlog>

            <whatlog bldinf=I{/path/to/src/path/to/group/bld.inf}' mmp='' config='foo.whatlog'>
                <bitmap>I{foo/bar}</bitmap>
            </whatlog>
        </buildinfo>

    createBuildData() outputs PlainBuildData objects from a set of INF paths, while
    raw data can be obtained using getFullInfo() and getFilePaths().
    '''
    def __init__(self):
        self.cdata = []
        self.cdString = ""
        self.infs = {}
        self.currentInfPath = ""
        self.currentArchive = {}
        self.logger = Logging.getLogger("pfw.datasources.whatlog")

    def characters(self, data):
        self.cdata.append(data)

    def startElement(self, tag, attributes):
        self.cdata = []
        if tag == "whatlog":
            self.currentInfPath = os.path.abspath(attributes.get("bldinf"))
            if self.currentInfPath not in self.infs:
                self.infs[self.currentInfPath] = []
        elif tag == "export":
            self.infs[self.currentInfPath].append({'type': 'file', 'path': attributes.get('destination'), 'fileType': tag})
        elif tag == "archive":
            self.currentArchive = {'type': 'archive', 'path': attributes.get("zipfile"), 'members': []}

    def endElement(self, tag):
        self.cdString = "".join(self.cdata).strip()
        if tag == "archive":
            self.infs[self.currentInfPath].append(self.currentArchive)
            self.currentArchive = {}
        elif tag == "member":
            self.currentArchive['members'].append(self.cdString)
        elif tag in ("resource", "build", "bitmap"):
            self.infs[self.currentInfPath].append({'type': 'file', 'path': self.cdString, 'fileType': tag})

    def getInfs(self):
        return self.infs.keys()

    def getFullInfo(self, inf):
        if inf not in self.infs:
            return
        for entry in self.infs[inf]:
            if entry["type"] == "file":
                yield entry
            elif entry["type"] == "archive":
                for path in entry["members"]:
                    yield {'type': 'file', 'path': path, 'fileType': "member", 'archive': entry["path"]}

    def getFilePaths(self, inf):
        if inf not in self.infs:
            self.logger.warning("No such inf %s"%inf)
            return
        for entry in self.infs[inf]:
            if entry["type"] == "file":
                yield entry["path"]
            elif entry["type"] == "archive":
                for path in entry["members"]:
                    yield path

    def createBuildData(self, infs, name=None, version=None, epocRoot=None):
        '''
        Create a PlainBuildData object from one or more infs.

        @param infs: Absolute paths to INF files
        @type infs: List
        @param name: Component name
        @type name: String
        @param version: Component version
        @type version: String
        @param epocRoot: Absolute path to epocroot. If not specified it will be
        deduced from the paths.
        @type epocRoot: String
        '''
        if not infs:
            raise ValueError, "Require at least one inf"
        bd = PlainBuildData()
        if epocRoot:
            bd.targetRoot = epocRoot
        if name:
            bd.componentName = name
        if version:
            bd.componentVersion = version
        for inf in infs:
            for absPath in self.getFilePaths(inf):
                absPath = os.path.abspath(absPath)
                path = ""
                if epocRoot and absPath.startswith(epocRoot):
                    path = absPath[len(epocRoot):].lstrip(os.sep)
                else:
                    if "epoc32" not in absPath:
                        self.logger.warning("Ignoring file - unable to deduce epocRoot - 'epoc32' not in file path '%s'." % absPath)
                        continue
                    epocRoot = absPath[:absPath.index("epoc32")]
                    path = absPath[absPath.index("epoc32"):]
                bd.addTargetFiles([path])
        return bd

    @staticmethod
    def getBuildData(source, epocRoot=None):
        '''
        Create a dictionary of BuildData objects with INFs as keys, one
        BuildData per INF.

        @param source: A file path or file-like object containing generic
        whatlog data.
        @type source: String or file-like object
        @return: Dictionary with INF:BuildData pairs
        '''
        reader = WhatLogReader()
        p = xml.sax.make_parser()
        p.setContentHandler(reader)
        if hasattr(source, "read"):
            file = source
        else:
            file = open(source)
        while True:
            data = file.read(1000000)
            if not data:
                break
            p.feed(data)
        p.close()
        infsAndBuildData = {}
        for inf in reader.getInfs():
            infsAndBuildData[inf] = reader.createBuildData(infs=[inf], epocRoot=epocRoot)
        return infsAndBuildData

class GenericLogReader(xml.sax.ContentHandler):
    '''
    Read output from whatlogs transformed using buildLogToGeneric.xsl.

    The format is::

        <components>
          <component bldinf="/path/to/src/path/to/bld.inf" config="tools2_deb">
              <file type="export">/path/to/epoc32/foo/bar</file>
          </component>
        </components>

    Where I{type} is xpath:whatlog/node() name and I{config} is xpath:whatlog@config.

    createBuildData() outputs PlainBuildData objects from a set of INF paths, while
    raw data can be obtained using getFullInfo() and getFilePaths().
    '''

    defaultStyleSheet = os.path.join(os.path.dirname(__file__), "buildLogToGeneric.xsl")

    def __init__(self):
        self.cdata = []
        self.cdString = ""
        self.logger = Logging.getLogger("pfw.datasources.genericlog")
        self.components = {}
        self.bldinf = ""
        self.config = ""
        self.type = ""

    def characters(self, data):
        self.cdata.append(data)

    def startElement(self, tag, attributes):
        self.cdata = []
        if tag == "component":
            self.bldinf = os.path.abspath(attributes.get("bldinf"))
            self.config = attributes.get("config")
            if self.bldinf not in self.components:
                self.components[self.bldinf] = []
        elif tag == "file":
            self.type = attributes.get("type")

    def endElement(self, tag):
        self.cdString = "".join(self.cdata).strip()
        if tag == "component":
            self.bldinf = ""
            self.config = ""
        elif tag == "file":
            self.components[self.bldinf].append({'type': self.type, 'path': self.cdString})

    def getInfs(self):
        return self.components.keys()

    def getFullInfo(self, inf):
        return self.components.get(inf)

    def getFilePaths(self, inf):
        if inf not in self.components:
            self.logger.warning("No such inf %s"%inf)
            return
        return [ f["path"] for f in self.components[inf] ]

    def createBuildData(self, infs, name=None, version=None, epocRoot=None):
        '''
        Create a PlainBuildData object from one or more infs.

        @param infs: Absolute paths to INF files
        @type infs: List
        @param name: Component name
        @type name: String
        @param version: Component version
        @type version: String
        @param epocRoot: Absolute path to epocroot. If not specified it will be
        deduced from the paths.
        @type epocRoot: String
        '''
        if not infs:
            raise ValueError, "Require at least one inf"
        bd = PlainBuildData()
        if epocRoot:
            bd.targetRoot = epocRoot
        if name:
            bd.componentName = name
        if version:
            bd.componentVersion = version
        for inf in infs:
            for absPath in self.getFilePaths(inf):
                absPath = os.path.abspath(absPath)
                if epocRoot and absPath.startswith(epocRoot):
                    path = absPath[len(epocRoot):].lstrip(os.sep)
                else:
                    if "epoc32" not in absPath:
                        self.logger.warning("Ignoring file - unable to deduce epocRoot - 'epoc32' not in file path '%s'." % absPath)
                        continue
                    epocRoot = absPath[:absPath.index("epoc32")]
                    path = absPath[absPath.index("epoc32"):]
                bd.addTargetFiles([path])
        return bd

    @staticmethod
    def getBuildData(source):
        '''
        Create a dictionary of BuildData objects with INFs as keys, one
        BuildData per INF.

        @param source: A file path or file-like object containing generic
        whatlog data.
        @type source: String or file-like object
        '''
        reader = GenericLogReader()
        p = xml.sax.make_parser()
        p.setContentHandler(reader)
        if hasattr(source, "read"):
            file = source
        else:
            file = open(source)
        while True:
            data = file.read(1000000)
            if not data:
                break
            p.feed(data)
        p.close()
        infsAndBuildData = {}
        for inf in reader.getInfs():
            infsAndBuildData[inf] = reader.createBuildData(infs=[inf])
        return infsAndBuildData

if __name__ == "__main__":
    #
    # An example of how to use WhatLog.
    #
    # - Parse Symbian build log directly using WhatLogReader:
    #
    #   WhatLog.py [-p|--print] logfile [inf inf2 ...]
    #
    # - Parse generic build log using GenericLogReader:
    #
    #   WhatLog.py -g|--generic [-p|--print] logfile [inf inf2 ...]
    #
    # - Convert Symbian build log using XSLT then parse using GenericLogReader
    #
    #   WhatLog.py -s|--stylesheet stylesheet.xsl [-p|--print] logfile [inf inf2 ...]
    #
    import logging
    from getopt import getopt
    from Ft.Xml.Xslt import Transform
    from Blocks.Packaging.BuildData import BuildData

    opts, leftovers = getopt(sys.argv[1:], "hvs:pg", ["stylesheet=", "print", "generic"])

    stylesheet = None
    isGeneric = False
    printInfs = False
    loglevel = logging.WARNING

    for k, v in opts:
        if k == "-v":
            loglevel = logging.DEBUG

    logging.basicConfig(level=loglevel)

    for k, v in opts:
        if k in ("-s", "--stylesheet"):
            stylesheet = v
        elif k in ("-p", "--print"):
            printInfs = True
        elif k in ("-g", "--generic"):
            isGeneric = True

    if leftovers:
        p = xml.sax.make_parser()
        if stylesheet:
            logging.debug("Using stylesheet %s", stylesheet)
            i = GenericLogReader()
            p.setContentHandler(i)
            p.feed(Transform(leftovers[0], stylesheet))
        else:
            if isGeneric:
                i = GenericLogReader()
                logging.debug("Reading generic XML")
            else:
                i = WhatLogReader()
                logging.debug("Reading Symbian XML")
            p.setContentHandler(i)
            f = open(leftovers[0])
            while True:
                data = f.read(1000000)
                if not data:
                    break
                p.feed(data)
        p.close()

        infs = i.getInfs()
        infs.sort()
        if printInfs:
            for inf in infs:
                print "%s" % inf
                for path in i.getFilePaths(inf):
                    print "\t", path

        # create a BuildData object from a list of INF paths
        if len(leftovers) > 1:
            bd = i.createBuildData("testcomponent", "1", leftovers[1:], "foo")
            assert isinstance(bd, BuildData), "not a build data object"
            assert bd.getComponentVersion() == "1", "bad version"
