#
# Copyright (c) 2010 Nokia Corporation and/or its subsidiary(-ies).
# All rights reserved.
# This component and the accompanying materials are made available
# under the terms of "Eclipse Public License v1.0"
# which accompanies this distribution, and is available
# at the URL "http://www.eclipse.org/legal/epl-v10.html".
#
# Initial Contributors:
# Nokia Corporation - initial contribution.
#
# Contributors:
#
# Description:
# File mapping
#

'''
FileMap is a filename -> owner package lookup table.  It stores minimal information about
the package owning a file. Packager compiles a single large FileMap of all files
in a release, enabling it to resolve dependencies between packages from
file->file or package->file dependencies.
'''

from cPickle import dump, load

from Blocks.Packaging.PackageWriter import VersionInfo, FileSet, Packageable
from Blocks.Packaging.Logging import Logging
from Blocks.Packaging import PackagingError

class MissingPackage(PackagingError):
    ''' No such package in filemap '''

class DuplicateFile(PackagingError):
    ''' The path already exists in the filemap '''

class DuplicatePackage(PackagingError):
    ''' The package already exists in the filemap '''

def externalsources(fn):
    '''
    A decorator to use external data sources if the filemap does not have
    the requested data.
    '''
    def wrapper(self, *args):
        out = fn(self, *args)
        if out:
            return out
        for source in self._externalSources:
            method = getattr(source, fn.func_name, None)
            if method:
                return method(*args)
        return None
    return wrapper

class StubPackage(VersionInfo):
    ''' A way to store minimal Package data '''
    def __init__(self, package=None):
        if package:
            self._name = package.getName()
            self.setIdentifier(package.getIdentifier())
            self._apiVersion = package.getApiVersion() or None
            self._apiFingerprint = package.getApiFingerprint() or None
        else:
            self._name = None
            self._apiVersion = None
            self._apiFingerprint = None

    def getName(self):
        return self._name

    def getApiVersion(self):
        return self._apiVersion

    def getApiFingerprint(self):
        return self._apiFingerprint

    def __str__(self):
        return "name: %s\ncomponentName: %s\napiVersion: %s\napiFingerprint: %s\nversion: %s\nbuild: %s\n" % (
            self.getName(),
            self.getComponentName(),
            self.getApiVersion(),
            self.getApiFingerprint(),
            self.getVersion(),
            self.getBuild())

class FileMap(object):
    ''' Contains file -> package mappings. '''
    def __init__(self, *args):
        '''
        Store the package name as well as version and API data of the package
        each file belongs to.

        @param args: A component or package from which to store the filenames
        @type args: [Packageable, FileSet or FileMap]
        '''
        self.files = {}
        self.stubs = {}
        self._externalSources = []
        for arg in args:
            self.addAny(arg)

    @staticmethod
    def getLogger():
        return Logging.getLogger("pfw.filemap")

    def registerExternalSource(self, source):
        self._externalSources.append(source)

    def addAny(self, any):
        '''
        Add files to FileMap
        @param any: Something that contains file and package information
        @type any: Packageable, FileSet or FileMap
        '''
        if isinstance(any, FileSet):
            self._addPackage(any)
        elif isinstance(any, Packageable):
            for package in any.getPackages():
                self._addPackage(package)
        elif isinstance(any, FileMap):
            for name in any.getPackageNames():
                if name in self.stubs:
                    raise DuplicatePackage("Will not add from Filemap containing data for package %s, which already exists.")

            for path in any.getPaths():
                try:
                    self._addPath(path, any.getPackageName(path))
                except MissingPackage:
                    stub = any.getPackage(path)
                    self.stubs[stub.getName()] = stub
                    self._addPath(path, any.getPackageName(path))
        else:
            raise TypeError("Unable to add this type to FileMap: '%s'" % type(any))

    def _addPackage(self, package):
        '''
        @param package: Package to add
        @type package: FileSet
        '''
        if package.getName() in self.stubs:
            raise DuplicatePackage, "Will not add Package %s, already exists."
        for file in package.getFiles():
            try:
                self._addPath(file.path, package.getName())
            except MissingPackage:
                self.stubs[package.getName()] = StubPackage(package)
                self._addPath(file.path, package.getName())

    def _addPath(self, path, packageName):
        assert isinstance(path, basestring)
        if path in self.files:
            self.getLogger().warning("Non-unique file %s, not adding to filemap. Already a member of %s", path, self.getPackageName(path))
        elif packageName not in self.stubs:
            raise MissingPackage, "Add stub for package %s before adding files to it."
        else:
            self.files[path] = packageName

    def getPaths(self):
        return self.files.__iter__()

    def getPackageNames(self):
        return self.stubs.__iter__()

    @externalsources
    def getPackage(self, filename):
        return self.stubs.get(self.getPackageName(filename))

    @externalsources
    def getPackageName(self, filename):
        '''
        @param filename: filename
        @type filename: String
        @rtype: String or None
        '''
        return self.files.get(filename)

#    @externalsources
#    def getComponentName(self, filename):
#        if filename in self._filemap:
#            return self._filemap[filename].getComponentName()
#        return None

    @externalsources
    def getPackageVersion(self, filename):
        if filename in self.files:
            return self.getPackage(filename).getVersion()

    @externalsources
    def getPackageBuild(self, filename):
        if filename in self.files:
            return self.getPackage(filename).getBuild()

    @externalsources
    def getPackageApiVersion(self, filename):
        if filename in self.files:
            return self.getPackage(filename).getApiVersion()

    @externalsources
    def getPackageApiFingerprint(self, filename):
        if filename in self.files:
            return self.getPackage(filename).getApiFingerprint()

    def loadFragment(self, filename):
        '''
        Appends previously pickled FileMap to this.

        @param filename: The filename or file-like object of a pickled FileMap
        @type filename: String
        '''
        self.addAny(FileMap.load(filename))

    @staticmethod
    def load(f):
        '''
        Read a previously pickled FileMap from a file.

        @param f: filename or file-like object
        @return: A FileMap loaded from f
        @rtype: FileMap
        '''
        if not hasattr(f, "read"):
            try:
                f = open(f, "r")
            except Exception:
                raise ValueError("not openable: '%s'" % f)
        f.seek(0)
        return load(f)

    def dump(self, f):
        '''
        Dump the FileMap using pickle

        @param f: filename or file-like object
        @type f: String
        '''
        if isinstance(f, basestring):
            f = open(f, "wb")
        dump(self, f, 0)