#
# Copyright (c) 2010 Nokia Corporation and/or its subsidiary(-ies).
# All rights reserved.
# This component and the accompanying materials are made available
# under the terms of "Eclipse Public License v1.0"
# which accompanies this distribution, and is available
# at the URL "http://www.eclipse.org/legal/epl-v10.html".
#
# Initial Contributors:
# Nokia Corporation - initial contribution.
#
# Contributors:
#
# Description:
# Component metadata storage
#

''' Component metadata storage '''

import time
import os

from Blocks.Packaging.FileMapping import FileMap
from Blocks.Packaging import PackagingError
from Blocks.singleinstance import SingleInstance

class IncompleteData(PackagingError):
    ''' There was a previous build but the requested data was not saved. '''

class NoPreviousBuild(PackagingError):
    ''' There have been no builds for the component. '''

class InternalError(PackagingError):
    ''' Internal problem '''

class PackagerStorage(object):
    '''Interface for providing the paths to metadata files required by a Packager.

    When instantiated, the build ID is incremented (if it's stored across uses).
    '''

    def __init__(self, *args, **kwargs):
        pass

    def writeFileMap(self, component):
        '''Write filemap to storage.

        @param component: A component
        @type component: Blocks.Packaging.PackageModel.Component
        '''
        raise NotImplementedError

    def writeMetaData(self, component):
        '''Write metadata to storage.

        @param component: A component
        @type component: Blocks.Packaging.PackageModel.Component
        '''
        raise NotImplementedError

    def getLastFileMapFile(self, componentName):
        '''Get path to file map file of most recent build.

        @param componentName: Name of the component
        @type componentName: String
        @return: Path
        @rtype: String
        '''
        raise NotImplementedError

    def getLastMetaDataFile(self, componentName):
        '''Get path to metadata file of most recent build.

        @param componentName: Name of the component
        @type componentName: String
        @return: Path
        @rtype: String
        '''
        raise NotImplementedError

    def getBuildId(self):
        '''Return the ID of the current build

        return: The ID
        rtype: String
        '''
        raise NotImplementedError

    def getComponentNames(self):
        '''Return a list of components whose metadata is found in this storage
        instance; this can be either metadata, files2packages or both.

        @return: Names of components
        @rtype: List(String)
        '''
        raise NotImplementedError

    def __str__(self):
        return self.__class__.__name__

class DefaultPackagerStorage(PackagerStorage):
    '''
    Store metadata in a directory hierarchy::

        + Basedir
        |
        -- + Component
           |
           -- Version

    A running build ID is maintained, but it can be overridden.
    '''

    def __init__(self, basedir, build=None, lock=False):
        '''
        @param basedir: The directory containing a directory for each component's metadata
        @type basedir: string
        @param build: Overrides the incrementing build ID
        @type build: int
        @param lock: Lock the storage for exclusive access by this instance.
        The lock remains until the instance is deleted or release() is called.
        @type lock: bool
        '''
        if not os.path.isdir(basedir):
            raise ValueError("Base directory '%s' does not exist. Cannot continue." % basedir)

        self._baseDir = basedir
        self._idFile = os.path.join(basedir, "build_id")
        self._lockDir = os.path.join(basedir, ".lock")

        try:
            self._lastBuild = self._readId()
        except NoPreviousBuild:
            self._lastBuild = 0

        if build:
            if int(build) < int(self._lastBuild):
                raise ValueError("Build ID (%s) must be equal or greater than the previous one (%s)" % (build, self._lastBuild))
            self._build = str(build)
            self._lastBuild = str(int(self._build) - 1)
        else:
            self._build = str(int(self._lastBuild) + 1)

        if lock:
            si = SingleInstance(basedir, True, ".si-lock")
            si.waitRelease()

    def writeFileMap(self, component):
        _mkdir(os.path.join(self._baseDir, component.getName()))
        if _mkdir(os.path.join(self._baseDir, component.getName(), self._build)):
            self._writeId()
        fMap = FileMap(component)
        fMap.dump(os.path.join(self._baseDir, component.getName(), self._build, "file2pkg.dat"))

    def writeMetaData(self, component):
        _mkdir(os.path.join(self._baseDir, component.getName()))
        if _mkdir(os.path.join(self._baseDir, component.getName(), self._build)):
            self._writeId()
        component.dump(os.path.join(self._baseDir, component.getName(), self._build, "meta.dat"))

    def getLastFileMapFile(self, componentName):
        if self._lastBuild == 0:
            raise NoPreviousBuild("No previous builds")
        last = os.path.join(self._baseDir, componentName, self._latestBuildFor(componentName), "file2pkg.dat")
        if not os.path.isfile(last):
            raise IncompleteData("No file mapping data was stored for the last build of the component; the expected file %s is does not exist." % last)
        else:
            return last

    def getLastMetaDataFile(self, componentName):
        if self._lastBuild == 0:
            raise NoPreviousBuild("No previous builds")
        last = os.path.join(self._baseDir, componentName, self._latestBuildFor(componentName, maxId=self._lastBuild), "meta.dat")
        if not os.path.isfile(last):
            raise IncompleteData("No metadata was stored for the last build of the component; the expected file %s is does not exist." % last)
        else:
            return last

    def getBuildId(self):
        return self._build

    def getComponentNames(self):
        return [f for f in os.listdir(self._baseDir) if f not in ("build_id", ".lock", ".si-lock")]

    def _writeId(self, id=None):
        '''
        Use lock to make sure only one instance will update the ID file. This is
        needed by multiprocess packager, it's not for running multiple
        Packager / PFW instances in parallel.
        '''
        id = id or self._build
        maxTries = 50
        wait = 0.01
        for f in range(maxTries):
            try:
                os.mkdir(self._lockDir)
                try:
                    try:
                        f = open(self._idFile, "wb")
                        f.write(id)
                        f.close()
                        return True
                    except Exception, e:
                        raise PackagerStorage.InternalError("Failed to update %s to ID %s:%s" % (self._idFile, self._build, e))
                finally:
                    os.rmdir(self._lockDir)
            except OSError, e:
                if e.errno != 17:
                    raise
                else:
                    time.sleep(wait)
                    continue
        # timeout, assume something has broken previously
        try:
            f = open(self._idFile, "wb")
            f.write(self._build)
            f.flush()
            f.close()
            return True
        except Exception:
            pass
        os.rmdir(self._lockDir)

    def _readId(self):
        ''' Get the last build from idFile, or raise NoPreviousBuild. '''
        maxTries = 50
        wait = 0.01
        for tries in range(maxTries):
            try:
                f = open(self._idFile)
                i = f.read().strip()
                assert i.isdigit(), "Corrupt ID file %s" % self._idFile
                f.close()
            except IOError, e:
                # build_id does not exist (yet?)
                # give it a chance
                if e.errno == 2:
                    time.sleep(wait)
                    continue
                else:
                    raise NoPreviousBuild, str(e)
            except Exception, e:
                raise NoPreviousBuild, str(e)
            return i
        raise NoPreviousBuild

    def _latestBuildFor(self, componentName, maxId=None):
        '''Find the directory containing the latest metadata for componentName.

        @param componentName: Component name
        @type componentName: String
        @param maxId: The ID shall not be greater than maxId
        @type maxId: Integer
        @return: Path to the metadata directory
        @rtype: String
        '''
        compDir = os.path.join(self._baseDir, componentName)
        if not os.path.isdir(compDir):
            raise NoPreviousBuild, "No previous builds for %s" % componentName
        dirs = [ d for d in os.listdir(compDir) if \
                    d.isdigit() \
                    and os.path.isdir(os.path.join(compDir, d))
                ]
        if not dirs:
            raise NoPreviousBuild("No metadata in build history for %s" % componentName)
        dirs.sort(lambda x, y: cmp(int(x), int(y)))
        latest = dirs.pop()
        if maxId and latest > maxId:
            if dirs:
                return dirs.pop()
            else:
                raise NoPreviousBuild("Build history for %s only has entries newer than %s" % (componentName, maxId))
        else:
            return latest

    def __str__(self):
        return "%s at %s" % (self.__class__.__name__, self._baseDir)

class OneoffStorage(PackagerStorage):
    '''
    OneoffStorage is for a one-off build without a directory structure for
    consecutive builds. Previously created metadata can be used, although
    only for acquiring metadata and not for calculating the build ID.
    '''
    def __init__(self, basedir=None, build=1, previousMetaDir=None, flat=False):
        '''
        Basedir is the output directory, previousMetaDir contains meta.dat of
        previous build (optional)

        A subdirectory is created in basedir for each component, unless flat is
        True. It's not advisable to use the flat option if metadata or filemap
        for more than one component is being written.

        Note also that when using the flat option the metadata is not easily
        accessible using OneoffStorage later on.

        @param basedir:         Output directory - if None nothing is written.
        @type basedir:          String
        @param build:           Mandatory for OneoffStorage
        @param previousMetaDir: Set this path to the basedir of the last build
                                to read old metadata. Assumes flat was not used,
                                i.e. directories named after components are
                                found in previousMetaDir.
        @type previousMetaDir:  String
        @param flat:            Set to True to put metadata files directly in
                                the base directory; normally they are stored in
                                a directory named after the component.
        @type flat:             Boolean
        '''
        if not build:
            raise AttributeError("Must have a build ID for OneOffStorage")
        if previousMetaDir and not os.path.isdir(previousMetaDir):
            raise ValueError("Old metadata path %s is not a directory" % previousMetaDir)
        PackagerStorage.__init__(self)
        self._baseDir = basedir
        self._previousMetaDir = previousMetaDir
        self._build = str(build)
        self._flat = flat

    def __str__(self):
        return "%s with %s metadata directory" % (self.__class__.__name__, self._baseDir or "no")

    def writeFileMap(self, component):
        if self._baseDir:
            if self._flat:
                dir = self._baseDir
            else:
                dir = os.path.join(self._baseDir, component.getName())
                _mkdir(dir)
            fMap = FileMap(component)
            fMap.dump(os.path.join(dir, "file2pkg.dat"))

    def writeMetaData(self, component):
        if self._baseDir:
            if self._flat:
                dir = self._baseDir
            else:
                dir = os.path.join(self._baseDir, component.getName())
                _mkdir(dir)
            component.dump(os.path.join(dir, "meta.dat"))

    def getLastMetaDataFile(self, componentName):
        ''' Get last metadata file '''
        if not self._previousMetaDir:
            raise NoPreviousBuild("Cannot get last metadata - no previous build directory was given to OneOffStorage")
        if self._flat:
            componentName = ""
        path = os.path.join(self._previousMetaDir, componentName, "meta.dat")
        if os.path.isfile(path):
            return path
        else:
            if not self._flat and os.path.isdir(os.path.join(self._previousMetaDir, componentName)):
                raise IncompleteData("No metadata was stored for the last build; the expected file %s is does not exist." % path)
            else:
                raise NoPreviousBuild("The metadata file %s does not exist" % path)


    def getLastFileMapFile(self, componentName):
        ''' Get last map file '''
        if not self._previousMetaDir:
            raise NoPreviousBuild("Cannot get last filemap - no previous build directory was given to OneOffStorage")
        if self._flat:
            componentName = ""
        path = os.path.join(self._previousMetaDir, componentName, "file2pkg.dat")
        if os.path.isfile(path):
            return path
        else:
            if not self._flat and os.path.isdir(os.path.join(self._previousMetaDir, componentName)):
                raise IncompleteData("No filemapping was stored for the last build; the expected file %s is does not exist." % path)
            else:
                raise NoPreviousBuild("The filemap file %s does not exist" % path)

    def getBuildId(self):
        return self._build

    def getComponentNames(self):
        '''
        Only ever return anything if a previousMetaDir was provided. Any subdirectory
        with metadata or file2packages is assumed to be a component name.
        '''
        if self._previousMetaDir:
            return [f for f in os.listdir(self._previousMetaDir)
                    if os.path.isfile(os.path.join(self._previousMetaDir, f, "meta.dat"))
                    or os.path.isfile(os.path.join(self._previousMetaDir, f, "file2pkg.dat"))]
        else:
            return []

def _mkdir(path):
    try:
        os.mkdir(path)
        return True
    except OSError, e:
        if e.errno != 17:
            raise e
        else:
            return False